#!/usr/bin/env python
# coding: utf-8

import sys
import os

try:
    prtr_path = os.environ['PRAETOR']
    sys.path.append(prtr_path)
    import praetor_settings_user as praetor_settings
    from praetor import praetorTemplates

except:
    from praetor import praetor_settings, praetorTemplates
    # from praetor import praetor_settings
    # import praetor_settings

sys.path.append(praetor_settings.provenance_directory + os.environ['PRAETOR_pipeline_id'] + '/')


def escape_forbidden_characters(in_string):
    '''
    Escape forbidden characters in input string in accordance with PROV-N
    :param in_string: input string to escape characters within
    :return: modified in_string with escaped characters
    '''
    forbidden = "\\(=)|,-:;[]+"
    allowed_special = '/._'
    in_string = [x if x.isalnum() else '\\'+x if x in forbidden else x if x in allowed_special else '' for x in in_string]
    final_string = ''.join(in_string)
    return final_string


def createActivity(activity, agent):
    '''
    Creates the activity and agent string for the PROV-N code and the relations between them, corresponding to the
    function and python code, respectively
    :param activity: Name of function
    :param agent: Type of agent (typically software agent)
    :return: PROV-N format string for agent, activity and relations
    '''
    activity = escape_forbidden_characters(activity)
    agent = escape_forbidden_characters(agent)
    activityString = """
    activity(var:message, var:messageStartTime, var:messageEndTime, [prtr:activityName = 'var:activityName',
     prtr:hadMemoryUsage = 'var:memoryUsage', prtr:activitySource = 'var:moduleName' ])
    agent(var:lifeline, [prtr:modules = 'var:modules', prtr:python_version = 'var:python_version', prov:type='prtr:{1}'])

    wasAssociatedWith(var:message, var:lifeline, - , [])
    """.format(activity, agent)

    return activityString


def createActivityStarted(activity, agent):
    '''
    Creates the activity and agent string for the PROV-N code and the relations between them, corresponding to the
    function and python code, respectively. For the case when a function was started by another
    :param activity: Name of function
    :param agent: Type of agent (typically software agent)
    :return: PROV-N format string for agent, activity and relations
    '''

    activity = escape_forbidden_characters(activity)
    agent = escape_forbidden_characters(agent)
    activityString = """
    activity(var:message, var:messageStartTime, var:messageEndTime, [prtr:activityName = 'var:activityName',
     prtr:hadMemoryUsage = 'var:memoryUsage', prtr:activitySource = 'var:moduleName' ])
    activity(var:message2, var:message2StartTime, var:message2EndTime)
    agent(var:lifeline, [prtr:modules = 'var:modules', prtr:python_version = 'var:python_version', prov:type='prtr:{1}'])

    wasAssociatedWith(var:message, var:lifeline, - , [])
    wasStartedBy(var:message2, -, var:message, -)
    """.format(activity, agent)

    return activityString


def generateBundle(function, bundleStart, bundleEnd):
    inEntities = praetorTemplates.createInputEntities(function[1], function[0])
    activity = createActivity(function[0], 'SoftwareAgent')
    activityStarted = createActivityStarted(function[0], 'SoftwareAgent')
    outEntities = praetorTemplates.createOutputEntities(function[2])


    bundleList = [bundleStart, inEntities, activity, outEntities, bundleEnd]
    bundle = '\n'.join(bundleList)

    bundleList2 = [bundleStart, inEntities, activityStarted, outEntities, bundleEnd]
    bundle2 = '\n'.join(bundleList2)

    return bundle, bundle2


def createOpenTemplate():
    fileFunc = """
    def fileAccess(fileName, mode, user_ID, size, hash_value):
        metadata = 'metadata'
        return fileName
    """
    func = praetorTemplates.variableSeparation(fileFunc)
    funcList = [x for x in func.values()]
    funcList = funcList[0]
    bundle, bundleStarted = generateBundle(funcList, praetorTemplates.bundleStart,
                                                            praetorTemplates.bundleEnd)

    with open(praetor_settings.provenance_directory + os.environ['PRAETOR_pipeline_id'] + '/' +
              'templates/pythonBuiltinFileAccess_template.provn', 'w') as f:
        f.write(bundle)

    with open(praetor_settings.provenance_directory + os.environ['PRAETOR_pipeline_id'] + '/' +
              'templates/pythonBuiltinFileAccessstarted_template.provn', 'w') as f:
        f.write(bundleStarted)



import os
import json
from praetor import praetorTemplates


def template_from_json(json):

    template = '\nbundle exe:bundle1\n'

    entities = {key: value for key, value in json['var'].items() if 'input' in key}
    for x in range(int(len(entities) / 2)):
        ent_str = "\nentity(var:input{0}, [prov:value = 'var:input{0}value'])\nused(var:message, var:input{0}, -, [prov:role='exe:input_{0}'])\n".format(x)
        template += ent_str

    activity = '''\nactivity(var:message, var:messageStartTime, var:messageEndTime, [prtr:activityName = 'var:activityName',\nprtr:hadMemoryUsage = 'var:memoryUsage', prtr:activitySource = 'var:moduleName', prov:label="var:comment"  ])\n'''
    agent = "agent(var:lifeline, [prtr:modules = 'var:modules', prtr:python_version = 'var:python_version', prov:type='prtr:SoftwareAgent'])\nwasAssociatedWith(var:message, var:lifeline, - , [])\n"

    if 'message2' in list(json['var'].keys()):
        activity_2 = '\nactivity(var:message2, var:message2StartTime, var:message2EndTime)'
        started_relation = '\nwasStartedBy(var:message2, -, var:message, -)\n'
        template += activity
        template += activity_2
        template += agent
        template += started_relation
    else:
        template += activity
        template += agent

    entities = {key: value for key, value in json['var'].items() if 'output' in key}
    for x in range(int(len(entities) / 2)):
        ent_str = "\nentity(var:output{0}, [prov:value = 'var:output{0}value'])\nwasGeneratedBy(var:output{0}, var:message, -)\n".format(x)
        template += ent_str

    template += "\nendBundle"
    return template

def combine_prov(json_file, agent_json_file, template_dir, out_prov, commented_globals):
    '''
    Combines the output of praetor code to produce a singular provenance file.
    :param json_file: Bindings file name from praetor, usually named full_json.json
    :param agent_json_file: Bindings file name from praetor for agent data, usually named agent_json.json
    :param template_dir: Directory name where prov templates were generated
    :param out_prov: Name of the combined provenance file
    :return: One provenance file produced named out_prov which contains all information from the run
    '''
    template_list = os.listdir(template_dir)
    template_short = [x.replace('_template.provn', '') for x in template_list]
    template_dictionary = {}
    prefix_lines = []

    for name, temp in zip(template_short, template_list):
        with open(template_dir + temp, 'r') as f:
            data = f.readlines()
        template_string = [x for x in data if x not in ['document\n', 'endDocument']]
        prefix_lines.extend([item for item in template_string if item[0:7] == 'prefix '])
        refined_data = [item for item in template_string if item not in prefix_lines]
        big_string = ''.join(refined_data)
        template_dictionary[name] = big_string

    prefix_lines_set = list(set(prefix_lines))
    prefix_string = ''.join(prefix_lines_set)

    with open(agent_json_file, 'r') as f:
        for line in f:
            line = line.replace('\n', '')
            json_object = json.loads(line)
            agent_var = json_object['agent']['var']

    with open(out_prov, 'w') as f:
        f.write('document\n{}prefix urn_uuid <urn:uuid:>\nprefix run <http://example.org/>'.format(prefix_string))
    track_globals = False
    with open(json_file, 'r') as f:
        with open(out_prov, 'a') as out:
            for line in f:
                line = line.replace('\n', '')
                json_object = json.loads(line)
                name = next(iter(json_object))
                try:
                    template = template_dictionary[name]
                except KeyError:
                    template = template_from_json(json_object[name])
                    template_dictionary[name] = template
                field_names = json_object[name]['var'].keys()
                param_id = []
                param_value = []
                param_type = []
                param_label = []
                comment_labels = []
                comment_id = []
                for field_name in field_names:
                    if any([x in field_name for x in ['praetorglobalinput', 'praetorglobaloutput']]):
                        track_globals = True
                        if field_name in ['praetorglobalinput', 'praetorglobaloutput']:
                            glob_id = json_object[name]['var'][field_name][0]["@id"]
                        else:
                            param_label_single = json_object[name]['var'][field_name][0]['@label']
                            param_label_single = param_label_single[param_label_single.find(':')+1:]
                            if param_label_single in commented_globals:
                                comment_labels.append(json_object[name]['var'][field_name][0]['@value'])
                            else:
                                param_label.append(json_object[name]['var'][field_name][0]['@label'])
                                param_id.append(json_object[name]['var'][field_name][0]['@id'])
                                param_value.append(json_object[name]['var'][field_name][0]['@value'])
                                param_type.append(json_object[name]['var'][field_name][0]['@type'])

                    elif next(iter(json_object[name]['var'][field_name][0])) == '@id':
                        template = template.replace('var:{},'.format(field_name),
                                                    str(json_object[name]['var'][field_name][0]['@id']) + ',')

                    elif next(iter(json_object[name]['var'][field_name][0])) in ['@type', '@value']:
                        template = template.replace("'var:{}'".format(field_name),
                                                    '"{}" %% {}'.format(
                                                        json_object[name]['var'][field_name][0]['@value'].replace('\n', ''),
                                                        json_object[name]['var'][field_name][0]['@type']))
                        if field_name in ['messageStartTime', 'messageEndTime', 'message2StartTime', 'message2EndTime']:
                            template = template.replace("var:{}".format(field_name),
                                                        str(json_object[name]['var'][field_name][0]['@value']))
                if track_globals:
                    activity_id = json_object[name]['var']['message'][0]['@id']
                    glob_bund = praetorTemplates.globalBundle(param_label, param_value, param_id, param_type,
                                                              activity_id)
                    if len(comment_labels) == 0:
                        template = template.replace(', prov:label="var:comment"', '')
                    else:
                        comment_list = ['prov:label="{}" '.format(sing_com) for sing_com in comment_labels]
                        comment_string = ','.join(comment_list)
                        template = template.replace('prov:label="var:comment"', comment_string)
                    template = template.replace('endBundle', glob_bund + '\nendBundle')

                for field_name in agent_var.keys():
                    variable = str(agent_var[field_name][0]).replace('(', '')
                    variable = variable.replace(')', '')
                    if len(variable) > 0:
                        template = template.replace("'var:{}'".format(field_name), "'{}'".format(variable))
                    if field_name == 'lifeline':
                        template = template.replace("var:{},".format(field_name), "{},".format(variable))
                out.write(template)
            out.write('\nendDocument')







import os

memory = False  # record allocated RAM per code segment - True/False
wrap_open = False  # record provenance for pythons open function - True/False
build_prov = False  # merge all recorded provenance files - True/False
flatten_prov = False  # requires provtoolbox installation, removes duplicates and shrinks
global_tracking = False # option to record global variables throughout the code

provenance_directory = './' # location for generated provenance storage
modules = []  # list of python modules/functions to record provenance for - use the name assigned to it in python
blacklist = []  # list of function to not include provenance for
staticmethods = [] # leave blank unless there are complaints about static methods
json_file = 'full_json.json' # name of json file. No need to change, it's defined here as many files need access to it
run_dir = os.path.dirname(os.path.realpath(__file__)) # os.getcwd()
unique_string = '!!!!!!!!!' # identifies func start/end in logs, must not be present in the logs otherwise
max_value_length = 128 # maximum no og chars for a value before it is stored separately from the provenance

if provenance_directory[-1] != '/':
    provenance_directory += '/'





import subprocess

from neo4j import GraphDatabase
import pandas as pd
from prov.model import ProvDocument
from prov2neo.client import Client
from neotime import DateTime

class Neo4jConnection:

    def __init__(self, uri, user, pwd):
        self.__uri = uri
        self.__user = user
        self.__pwd = pwd
        self.__driver = None
        try:
            self.__driver = GraphDatabase.driver(self.__uri, auth=(self.__user, self.__pwd))
        except Exception as e:
            print("Failed to create the driver:", e)

    def close(self):
        if self.__driver is not None:
            self.__driver.close()

    def query(self, query, db=None):
        assert self.__driver is not None, "Driver not initialized!"
        session = None
        response = None
        try:
            session = self.__driver.session(database=db) if db is not None else self.__driver.session()
            response = list(session.run(query))
        except Exception as e:
            print("Query failed:", e)
        finally:
            if session is not None:
                session.close()
        return response


def create_conn(prov_file, add="localhost:7687", usr='neo4j', password='admin', name='neo4j', scheme='bolt',
                uri="bolt://localhost:7687"):
    '''
    Uploads provenance file to neo4j database and establishes a connection to it
    :param prov_file: Provenance file to upload
    :param add: Address of neo4j database
    :param usr: username for neo4j
    :param password: password for neo4j
    :param name: name of database to upload to
    :param scheme: uri scheme
    :param uri: uri to connect to neo4j
    :return: Connection to neo4j database to be used for queries
    '''
    if prov_file[-5:] != '.json':
        extension = prov_file[prov_file.find('.'):]
        json_file = prov_file.replace(extension, '.json')
        subprocess.call(['provconvert', '-infile', prov_file, '-outfile', json_file])
    else:
        json_file = prov_file

    with open(json_file, 'r') as f:
        data = f.read()
    data = data.replace('"prov:starter":', '"prov:trigger":')
    with open(json_file, 'w') as f:
        f.write(data)

    graph = ProvDocument.deserialize(source=json_file, format="json")

    client = Client()
    client.connect(
        address=add,
        user=usr,
        password=password,
        name=name,
        scheme=scheme
    )
    client.import_graph(graph)
    conn = Neo4jConnection(uri=uri, user=usr, pwd=password)

    return conn


def query_handler(conn, query_string):
    '''
    Passes queries to Neo4j
    :param conn: Connection to Neo4j
    :param query_string: Query to pass
    :return: query results in a pandas dataframe
    '''
    result = conn.query(query_string, db='neo4j')
    return result


def clear_data(conn):
    '''
    Deletes all data from target neo4j graph database
    :param conn: connection to target graph
    :return:
    '''
    conn.query('MATCH (n) DETACH DELETE n', db='neo4j')


def activity_query(conn):
    query = "MATCH (n:Activity) RETURN n"
    result = query_handler(conn, query)
    df = pd.DataFrame([dict(r) for record in result for r in record])
    return df


def output_query(conn):
    query = "MATCH (n) - [r:wasGeneratedBy] -> (p) RETURN n"
    result = query_handler(conn, query)
    df = pd.DataFrame([dict(r) for record in result for r in record])
    return df


def input_query(conn):
    query = "MATCH (n) - [r:used] -> (p) RETURN p"
    result = query_handler(conn, query)
    df = pd.DataFrame([dict(r) for record in result for r in record])
    return df


def memory_query(conn):
    query = '''
    MATCH (n) WITH AVG(toFloat(n.`prtr:hadMemoryUsage`)) AS average 
    MATCH (n), (r)
    WHERE toFloat(r.`prtr:hadMemoryUsage`) > average
    RETURN r
    '''
    result = query_handler(conn, query)
    df = pd.DataFrame([dict(r) for record in result for r in record])
    return df


def source_query(conn):
    query = "MATCH (n:Activity) WHERE (n.`prtr:activitySource` = 'main') RETURN n"
    result = query_handler(conn, query)
    df = pd.DataFrame([dict(r) for record in result for r in record])
    return df
